/*******************************************************************************
*    Robot "Giskard" 1.0.                                                      *
*    Copyright (C) 2009 Luca Bertoncello                                       *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    Circuits by Marian Neubert marian@tesla-crew.de                           *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>
#include <termios.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "inc/newcomm.h"

/**
 * Liest die Optionen vom Kommandozeile
 *
 * @param int            argc  Anzahl der Parameter im Kommandozeile
 * @param char*          argv  Die Parameter im Kommandozeile
 * @param struct options opts  Die Beschreibung der möglichen Parameter
 */
int getOptions(int argc, char* argv[], struct options opts)
{
  int l, k, found, invalidOption;
  char *opt, *par;
  char *tOptLong, tOptShort[3];

  for(invalidOption = k = 0; k < opts.nOptions; k++)
  {
    if(opts.opt[k].type == INT || opts.opt[k].type == BOOL)
    {
      if(opts.opt[k].defValue != NULL)
        *(int*)opts.opt[k].value = *(int*)opts.opt[k].defValue;
      else
        *(int*)opts.opt[k].value = 0;
    }
    if(opts.opt[k].type == STRING)
    {
      if(opts.opt[k].defValue != NULL)
        strcpy(opts.opt[k].value, opts.opt[k].defValue);
      else
        strcpy(opts.opt[k].value, "");
    }
  }
  for(l = 1; l < argc; l++)
  {
    if(argv[l][0] == '-')
    {
      if(strstr(argv[l], "=") != NULL)
      {
        opt = strtok(argv[l], "=");
        par = strtok(NULL, "=");
      }
      else
      {
        opt = argv[l];
        par = NULL;
      }
      for(found = k = 0; k < opts.nOptions; k++)
      {
        tOptLong = malloc(sizeof(char) * (strlen(opts.opt[k].optNameLong) + 2));
        sprintf(tOptLong, "--%s", opts.opt[k].optNameLong);
        sprintf(tOptShort, "-%c", opts.opt[k].optNameShort);
        if((strcmp(tOptLong, opt) == 0) || (strcmp(tOptShort, opt) == 0))
        {
          found = 1;
          if(opts.opt[k].type == INT)
          {
            if(par == NULL)
            {
              par = argv[l + 1];
              l++;
            }
            *(int*)opts.opt[k].value = atoi(par);
          }
          if(opts.opt[k].type == STRING)
          {
            if(par == NULL)
            {
              par = argv[l + 1];
              l++;
            }
            strcpy(opts.opt[k].value, par);
          }
          if(opts.opt[k].type == BOOL)
          {
            *(int*)opts.opt[k].value = 1;
          }
        }
        free(tOptLong);
      }
      if(found == 0)
      {
        fprintf(stderr, opts.invalidOptionMsg, opt);
        invalidOption++;
      }
    }
  }

  return(invalidOption);
}

/**
 * Startet das Programm als Daemon
 *
 * @param char* pidfile  Die Datei mit dem PID
 */
void daemonize(char *pidfile)
{
  pid_t pid, sid;
  FILE *fp;

  pid = fork();
  if (pid < 0)
    exit(EXIT_FAILURE);

  if (pid > 0)
  {
    fp = fopen(pidfile, "w");
    fprintf(fp, "%d\n", pid);
    fclose(fp);
    exit(EXIT_SUCCESS);
  }

  umask(0);

  sid = setsid();
  if (sid < 0)
    exit(EXIT_FAILURE);

  if ((chdir("/")) < 0)
    exit(EXIT_FAILURE);

  close(STDIN_FILENO);
  close(STDOUT_FILENO);
  close(STDERR_FILENO);
}

/**
 * Loggt was das Programm empfangen hat in die Datei logfile
 *
 * @param char* str     Die String zu loggen
 * @param int   level   Das Log-Level
 */
void logger(char *str, int level)
{
  struct tm *tl;
  time_t t;
  char timeStr[255];
  FILE *fp;

  t = time(NULL);
  tl = localtime(&t);
  sprintf(timeStr, "%02d.%02d.%4d %02d:%02d:%02d", tl->tm_mday, tl->tm_mon + 1, tl->tm_year + 1900, tl->tm_hour, tl->tm_min, tl->tm_sec);

  if(nodetach == TRUE)
    fp = stdout;
  else
    fp = fopen(logfile, "at");

  if(level == LEVELDEBUG && hasToDebug == 1)
    fprintf(fp, "%s [%s]: %s\n", timeStr, levelStr[level], str);
  else
    if(level == LEVELERROR || level == LEVELINFO)
      fprintf(fp, "%s [%s]: %s\n", timeStr, levelStr[level], str);

  if(nodetach == 0)
    fclose(fp);
}

/**
 * Öffnet die serielle Schnittstelle, anhand der
 * angegebenen Parametern
 *
 * @return   0 wenn alles in Ordnung, -1 im Falle vom Probleme
 */
int openPort()
{
  struct termios options;
  char err[255];
  int status, relay;

  fd = open(modemDevice, O_RDWR | O_NOCTTY | O_NONBLOCK);
  if (fd == -1)
  {
    sprintf(err, "Unable to open device [%s]", modemDevice);
    logger(err, LEVELERROR);
    return(-1);
  }
  else
  {
    tcgetattr(fd, &options);

    cfsetispeed(&options, baudRate);
    cfsetospeed(&options, baudRate);

    options.c_cflag |= (CLOCAL | CREAD);
    options.c_cflag &= ~PARENB;
    options.c_cflag &= ~CSTOPB;
    options.c_cflag &= ~CSIZE;
    options.c_cflag |= CS8;
    options.c_cflag &= ~CRTSCTS;
    options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
    options.c_iflag &= ~INLCR;                    // Wenn es nicht so ist, dann werden einige Daten FALSCH interpretiert!!!
    options.c_iflag &= ~ICRNL;                    // zum Beispiel, wenn man ein Packet mit Länge 10 schickt,
    options.c_oflag &= ~ONLCR;                    // es wird eine Länge 13 vom MicroController gelesen!
    options.c_oflag &= ~OCRNL;                    // Also, keine Übersetzung!!
    options.c_iflag &= ~(IXON | IXOFF | IXANY);   // Kein XON/XOFF, so daß nicht die Länge 19 als Länge 67 gelesen wird

    ioctl(fd, TIOCMGET, &status);
    if(status & TIOCM_RTS)
      relay = 0;
    else
      relay = 1;
    tcsetattr(fd, TCSANOW, &options);
    if(relay == 1)
      status |= TIOCM_RTS;
    else
      status &= ~TIOCM_RTS;
    ioctl(fd, TIOCMSET, &status);
  }
  return(0);
}

/**
 * Liest die Daten vom PIPE und serielle Schnittstelle.
 * Schickt die Daten dem µC und verwaltet die Antworten
 *
 * @param char* pidfile   Die Datei mit dem PID
 * @param int   nodetach  TRUE wenn das Programm in der Console laufen soll
 *                        FALSE wenn das Program als Daemon läuft
 */
void doWork(char *pidfile, int nodetach)
{
  char buf[255];
  int n, fifo;

  rs232QueueCount = 0;
  if((mkfifo(commfile, 0666) == -1) && (errno != EEXIST))
  {
    sprintf(buf, "Error creating the named pipe [%s]", commfile);
    logger(buf, LEVELERROR);
    quitFunc(0);
  }

  if(openPort() == 0)
  {
    signal(SIGINT, quitFunc);
    signal(SIGQUIT, quitFunc);
    signal(SIGTERM, quitFunc);

    rs232ErrorCount = 0;
    sprintf(buf, "Device: %s - BaudRate: %s", modemDevice, baud);
    logger(buf, LEVELINFO);
    logger("Starting...", LEVELINFO);
    fifo = open(commfile, O_RDONLY | O_NONBLOCK);
    while(!STOP_PROGRAM)
    {
      bzero(buf, 255);
      n = read(fifo, buf, 255); // Lese vom PIPE
      if(n > 0)                 // Wenn Daten, dann verwalten
        managePIPE(buf);
      n = read(fd, buf, 1);     // Lese vom RS232
      if(n == 1)                // Wenn Daten, dann verwalten
        manageRS232(buf[0]);
      flushQueue();
      usleep(250);              // 250µS Pause wegen hohen Last
    }
    close(fifo);
  }
  unlink(commfile);
  if(nodetach == 0)
    unlink(pidfile);
  exit(EXIT_SUCCESS);
}

/**
 * Trapfunktion für CTRL+c.
 * Endet das Programm.
 *
 * @param int signum    Wird ignoriert, muß aber angegeben werden
 */
void quitFunc(int signum)
{
  STOP_PROGRAM = TRUE;
  close(fd);
  logger("Quit", LEVELINFO);
}

/**
 * Prüft, ob die Parität der empfangenen Daten stimmt
 *
 * @param struct dataFromRS232 d  Die empfangenen Daten
 * @return                        TRUE wenn stimmt, FALS ansonsten
 */
boolean checkParity(struct dataFromRS232 d)
{
  int l;

  calculatedParity = d.deviceFrom + d.deviceTo + d.packetSize;
  for(l = 0; l < d.packetSize; l++)
    calculatedParity += d.data[l];
  calculatedParity %= 256;

  if(calculatedParity != d.parity)
    return FALSE;
  return TRUE;
}

/**
 * Rechnet die Parität des zu sendenden Pakets und fügt sie
 * in das Paket hinzu
 *
 * @param struct dataFromRS232 d  Die zu sendenden Daten
 */
void addParityToPacket(struct dataFromRS232 *d)
{
  int parity, l;

  parity = d->deviceFrom + d->deviceTo + d->packetSize;
  for(l = 0; l < d->packetSize; l++)
    parity += d->data[l];
  d->parity = parity % 256;
}

/**
 * Legt ein Datenpaket an
 *
 * @param int   fromDev           Absendergerät
 * @param int   toDev             Empfängergerät
 * @param char* d                 Die Daten
 * @return struct dataFromRS232   Das Datenpaket
 */
struct dataFromRS232 createData(int fromDev, int toDev, char *d)
{
  struct dataFromRS232 data;

  data.version        = MYPROTOCOLVERSION;
  data.deviceFrom     = fromDev;
  data.deviceTo       = toDev;
  data.packetSize     = strlen(d);
  bzero(data.data, MAXDATALENGTH);
  strcpy(data.data, d);
  addParityToPacket(&data);

  return data;
}

/**
 * Fügt die Daten in die Queue, um sie zu senden
 *
 * @param struct dataFromRS232 d  Die zu sendenden Daten
 * @return boolean                TRUE wenn alles in Ordnung, FALSE wenn Problemen
 */
boolean addDataToQueue(struct dataFromRS232 d)
{
  if(rs232QueueCount < MAXQUEUE)
    rs232Queue[rs232QueueCount++] = d;
  else
    return FALSE;
  return TRUE;
}

/**
 * Schickt die Daten, die in der Queue sind
 */
void flushQueue()
{
  int n;
  char buf[300];
  time_t t;

  if(wait_ACK == TRUE)
  {
    t = time(NULL);
    if((t - waitACKStartTime) > ACK_TIMEOUT)
    {
      logger("Timeout while waiting for ACK", LEVELDEBUG);
      wait_ACK = FALSE;
    }
  }
  if(rs232QueueCount > 0 && wait_ACK == FALSE)
  {
    sprintf(buf, "Send: %d|%d|%d|%d|%s", rs232Queue[0].version, rs232Queue[0].deviceFrom, rs232Queue[0].deviceTo, rs232Queue[0].packetSize, rs232Queue[0].data);
    logger(buf, LEVELDEBUG);
    sprintf(buf, "%c%c%c%c", (unsigned char) rs232Queue[0].version,
                             (unsigned char) rs232Queue[0].deviceFrom,
                             (unsigned char) rs232Queue[0].deviceTo,
                             (unsigned char) rs232Queue[0].packetSize);
    n = write(fd, buf, 4);

    n = write(fd, rs232Queue[0].data, rs232Queue[0].packetSize);
    if (n < rs232Queue[0].packetSize)
      logger("ERROR BY SENDING DATA!!!", LEVELERROR);

    sprintf(buf, "%c", (unsigned char) rs232Queue[0].parity);
    n = write(fd, buf, 1);
    if (n < 1)
      logger("ERROR BY SENDING PARITY!!!", LEVELERROR);

    wait_ACK = TRUE;
    waitACKStartTime = time(NULL);
  }
}

/**
 * Schickt ein ACK
 */
void sendSigACK()
{
  int n;
  char str[2];

  str[0] = ACK;
  n = write(fd, str, 1);
  if (n < 1)
    logger("ERROR BY SENDING ACK!!!", LEVELERROR);
}

/**
 * Schickt ein NACK
 */
void sendSigNACK()
{
  int n;
  char str[2];

  str[0] = NACK;
  n = write(fd, str, 1);
  if (n < 1)
    logger("ERROR BY SENDING NACK!!!", LEVELERROR);
}

/**
 * Schickt ein PR_MISMATCH
 */
void sendSigPRMismatch()
{
  int n;
  char str[2];

  str[0] = PR_MISMATCH;
  n = write(fd, str, 1);
  if (n < 1)
    logger("ERROR BY SENDING PR_MISMATCH!!!", LEVELERROR);
}

/**
 * Verwaltet die Daten vom PIPE an das µC
 *
 * @param char* buf   Die Daten, in dem Format:
 *                      Absender|Empfänger|Daten
 */
void managePIPE(char *buf)
{
  int from, to;
  char data[256];
  char str[1024];
  struct dataFromRS232 d;

  buf[strlen(buf) - 1] = 0;
  sscanf(buf, "%d|%d|%[^\"]", &from, &to, data);
  if(strlen(data) < MAXDATALENGTH)
  {
    d = createData(from, to, data);
    sprintf(str, "Send: From [%d] To [%d] Data [%s] Parity [%02X]", from, to, data, d.parity);
    logger(str, LEVELINFO);
    if(addDataToQueue(d))
      logger("Added to the queue", LEVELINFO);
    else
      logger("Problem by adding to the queue", LEVELERROR);
  }
  else
    logger("Data too long! Not transmitted", LEVELERROR);
}

/**
 * Holt n Bytes vom seriellen Schnittstelle
 *
 * @param int   fd    Der FileDescriptor der TTY
 * @param char* buf   Puffer, wo die Daten gespeichert werden
 * @param int   n     Anzahl der zu empfangenen Bytes
 * @return int        Anzahl der empfangenen Bytes
 */
int getNBytes(int fd, char *buf, int n)
{
  int c, l, stop = 0;
  char b[2];
  time_t t;

  for(l = 0; (l < n) && (stop == 0);)
  {
    c = read(fd, b, 1);
    if(c == 1)
      buf[l++] = b[0];
    t = time(NULL);
    if((t - lastStartTime) > TIMEOUT_SERIAL)
    {
      logger("Timeout from serial", LEVELDEBUG);
      stop = 1;
    }
  }

  return(l);
}

/**
 * Verwaltet die Daten vom µC. Die Funktion wird aufgerufen, wenn ein Byte vom TTY ankommt
 *
 * @param char firstByte    Der erste Byte
 */
void manageRS232(char firstByte)
{
  char buf[255], str[1024];
  int n, len;
  int manage = 0;
  struct dataFromRS232 data;

  logger("Signal from TTY", LEVELDEBUG);
  lastStartTime = time(NULL);
  bzero(buf, 255);
  if(wait_ACK)
  {
    wait_ACK = FALSE;
    sprintf(str, "ACK from µC: [%02X] (%s)",
        (unsigned char) firstByte,
        ((unsigned char) firstByte) == ACK ? "ACK" :
          ((unsigned char) firstByte) == NACK ? "NACK" : "PR_MISMATCH");
    logger(str, LEVELDEBUG);
    if(((unsigned char) firstByte) == ACK)
    {
      for(n = 0; n < rs232QueueCount; n++)
        rs232Queue[n] = rs232Queue[n + 1];
      rs232QueueCount--;
      rs232ErrorCount = 0;
      logger("The recipient confirmed the data receive...", LEVELINFO);
    }
    else
    {
      if(((unsigned char) firstByte) == NACK)
        logger("ERROR BY DATA SEND!!!", LEVELERROR);
      else
        logger("PARITY ERROR ON SEND!!!", LEVELERROR);
      rs232ErrorCount++;
      if(rs232ErrorCount >= MAX_RETRIES)
      {
        for(n = 0; n < rs232QueueCount; n++)
          rs232Queue[n] = rs232Queue[n + 1];
        rs232QueueCount--;
        rs232ErrorCount = 0;
        logger("MAX RETRY ACHIEVED PACKET WILL BE DELETE FROM THE QUEUE!!!", LEVELERROR);
      }
    }
  }
  else
  {
    data.version = firstByte;
    n = getNBytes(fd, buf, 3);
    sprintf(str, "n: %d - buf: [%d|%d|%d|%d]", (n + 1), data.version, buf[0], buf[1], buf[2]);
    logger(str, LEVELDEBUG);
    if(data.version != MYPROTOCOLVERSION)
      logger("VERSION NUMBERS ARE DIFFERENT!!", LEVELERROR);
    else
    {
      data.deviceFrom = buf[0];
      data.deviceTo = buf[1];
      data.packetSize = buf[2];
      if(data.packetSize == 0)
        len = 1;
      else
        len = data.packetSize;
      bzero(buf, 255);
      n = getNBytes(fd, buf, len);
      sprintf(str, "n: %d - len: %d", n, len);
      logger(str, LEVELDEBUG);
      if(n == len)
      {
        strcpy(data.data, buf);
        n = getNBytes(fd, buf, 1);
        data.parity = (unsigned char) buf[0];
        if(checkParity(data) == TRUE)
        {
          sendSigACK();
          manage = 1;
          logger("Data are confirmed with ACK", LEVELINFO);
        }
        else
        {
          sendSigPRMismatch();
          sprintf(str, "Parity error (received: %02X, calculated: %02X)! Data are NOT confirmed", data.parity, calculatedParity);
          logger(str, LEVELERROR);
        }
      }
      else
      {
        logger("ERROR BY RECEIVING DATA!!!", LEVELERROR);
        sendSigNACK();
        printData(data);
      }
    }
  }

  if(manage)
    printData(data);
}

/**
 * Loggt die empfangenen Daten vom µC
 *
 * @param struct dataFromRS232 data  Die Daten
 */
void printData(struct dataFromRS232 data)
{
  int l;
  FILE *fp;
  char str[1024], s[5];

  fp = fopen(answerfile, "wt");
  fprintf(fp, "%s\n", data.data);
  fclose(fp);
  logger("Received Data:", LEVELDEBUG);
  sprintf(str, "Version: %d", data.version);
  logger(str, LEVELDEBUG);
  sprintf(str, "From device: %d", data.deviceFrom);
  logger(str, LEVELDEBUG);
  sprintf(str, "To device: %d", data.deviceTo);
  logger(str, LEVELDEBUG);
  sprintf(str, "Packet size: %d", data.packetSize);
  logger(str, LEVELDEBUG);
  logger("Data:", LEVELDEBUG);
  sprintf(str, "  [%s]  ", data.data);
  logger(str, LEVELDEBUG);
  bzero(str, 1024);
  for(l = 0; l < data.packetSize; l++)
  {
    sprintf(s, "%02X ", (unsigned char) data.data[l]);
    strcat(str, s);
  }
  logger(str, LEVELDEBUG);
  sprintf(str, "Parity: %02X (%s)", data.parity, checkParity(data) ? "OK" : "PROBLEM");
  logger(str, LEVELDEBUG);
  logger("---------------", LEVELDEBUG);
}

/**
 * Zeigt die Hilfe über das Programm
 */
void printHelp()
{
  printf("%s %s\n\n", NAME, VERSION);
  printf("Usage:\n");
  printf("%s [options]:\n", NAME);
  printf("  --help, -h           Display this help\n");
  printf("  --version, -v        Display version number\n");
  printf("  --verbose, -V        Increase verbosity\n");
  printf("  --nodetach, -D       Don't run as daemon\n");
  printf("  --pidfile <file>     Save the PID in <file>. Default: %s\n", DEFAULTPIDFILE);
  printf("  --logfile <file>     Logs in <file>. Default: %s\n", DEFAULTLOGFILE);
  printf("  --device <device>    Use <device> for the communication with the MicroController. Default: %s\n", DEFAULTMODEMDEVICE);
  printf("  --baudrate <rate>    Send/receive data with BaudRate <rate>. Default: %s\n", DEFAULTBAUDRATE);
  printf("  --commfile <file>    Use <file> to receive data for Microcontroller. Default: %s\n", DEFAULTCOMMFILE);
  printf("  --answerfile <file>  Use <file> to send data from MicroController. Default: %s\n", DEFAULTANSWERFILE);
  exit(0);
}

/**
 * Setzt die Geschwindigkeit der RS232
 *
 * @param char* baud    Die Geschwindigkeit als String.
 *                      Mögliche Werte: 9600, 19200, 38400, 57600, 115200
 * @return int          0 wenn alles in Ordnung, -1 falls Problemen
 */
int setBaudRate(char *baud)
{
  baudRate = B0;
  if(strcmp(baud, "9600") == 0)
    baudRate = B9600;
  if(strcmp(baud, "19200") == 0)
    baudRate = B19200;
  if(strcmp(baud, "38400") == 0)
    baudRate = B38400;
  if(strcmp(baud, "57600") == 0)
    baudRate = B57600;
  if(strcmp(baud, "115200") == 0)
    baudRate = B115200;

  if(baudRate == B0)
  {
    fprintf(stderr, "Valid Baudrate are:\n  - 9600\n  - 19200\n  - 38400\n  - 57600\n  - 115200\n");
    return(-1);
  }
  return(0);
}

/**
 * Main...
 */
int main(int argc, char* argv[])
{
  struct options opts;
  int help, version, l;
  char defPid[255], defLog[255], defTty[255], defComm[255], defAnswer[255], defBaud[255];

  strcpy(defPid, DEFAULTPIDFILE);
  strcpy(defLog, DEFAULTLOGFILE);
  strcpy(defTty, DEFAULTMODEMDEVICE);
  strcpy(defComm, DEFAULTCOMMFILE);
  strcpy(defAnswer, DEFAULTANSWERFILE);
  strcpy(defBaud, DEFAULTBAUDRATE);

  opts.nOptions = 10;
  opts.invalidOptionMsg = malloc(sizeof(char) * 255);
  strcpy(opts.invalidOptionMsg, "Invalid option [%s]\n");
  opts.opt = malloc(sizeof(struct option) * opts.nOptions);
  for(l = 0; l < opts.nOptions; l++)
    opts.opt[l].optNameLong = malloc(sizeof(char) * 20);
  strcpy(opts.opt[0].optNameLong, "help");
  opts.opt[0].optNameShort = 'h';
  opts.opt[0].type = BOOL;
  opts.opt[0].value = &help;
  opts.opt[0].defValue = NULL;
  strcpy(opts.opt[1].optNameLong, "verbose");
  opts.opt[1].optNameShort = 'V';
  opts.opt[1].type = BOOL;
  opts.opt[1].value = &hasToDebug;
  opts.opt[1].defValue = NULL;
  strcpy(opts.opt[2].optNameLong, "version");
  opts.opt[2].optNameShort = 'v';
  opts.opt[2].type = BOOL;
  opts.opt[2].value = &version;
  opts.opt[2].defValue = NULL;
  strcpy(opts.opt[3].optNameLong, "nodetach");
  opts.opt[3].optNameShort = 'D';
  opts.opt[3].type = BOOL;
  opts.opt[3].value = &nodetach;
  opts.opt[3].defValue = NULL;
  strcpy(opts.opt[4].optNameLong, "pidfile");
  opts.opt[4].optNameShort = 0;
  opts.opt[4].type = STRING;
  opts.opt[4].value = pidfile;
  opts.opt[4].defValue = defPid;
  strcpy(opts.opt[5].optNameLong, "logfile");
  opts.opt[5].optNameShort = 0;
  opts.opt[5].type = STRING;
  opts.opt[5].value = logfile;
  opts.opt[5].defValue = defLog;
  strcpy(opts.opt[6].optNameLong, "device");
  opts.opt[6].optNameShort = 0;
  opts.opt[6].type = STRING;
  opts.opt[6].value = modemDevice;
  opts.opt[6].defValue = defTty;
  strcpy(opts.opt[7].optNameLong, "baudrate");
  opts.opt[7].optNameShort = 0;
  opts.opt[7].type = STRING;
  opts.opt[7].value = baud;
  opts.opt[7].defValue = defBaud;
  strcpy(opts.opt[8].optNameLong, "commfile");
  opts.opt[8].optNameShort = 0;
  opts.opt[8].type = STRING;
  opts.opt[8].value = commfile;
  opts.opt[8].defValue = defComm;
  strcpy(opts.opt[9].optNameLong, "answerfile");
  opts.opt[9].optNameShort = 0;
  opts.opt[9].type = STRING;
  opts.opt[9].value = answerfile;
  opts.opt[9].defValue = defAnswer;

  if(getOptions(argc, argv, opts) != 0 || help != 0)
    printHelp();

  if(version == 1)
    printf("%s %s\n", NAME, VERSION);
  if(setBaudRate(baud) == 0)
  {
    if(nodetach == FALSE)
      daemonize(pidfile);
    doWork(pidfile, nodetach);
  }
  return 0;
}
